/*
 * @copyright Copyright (c) Open-Xchange GmbH, Germany <info@open-xchange.com>
 * @license AGPL-3.0
 *
 * This code is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with OX App Suite.  If not, see <https://www.gnu.org/licenses/agpl-3.0.txt>.
 *
 * Any use of the work other than as authorized under this license or copyright law is prohibited.
 *
 */

package com.openexchange.authentication.application.impl.api;

import java.util.List;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import com.openexchange.ajax.requesthandler.AJAXRequestData;
import com.openexchange.ajax.requesthandler.AJAXRequestResult;
import com.openexchange.ajax.requesthandler.Converter;
import com.openexchange.ajax.requesthandler.ResultConverter;
import com.openexchange.authentication.application.AppPasswordApplication;
import com.openexchange.exception.OXException;
import com.openexchange.tools.servlet.OXJSONExceptionCodes;
import com.openexchange.tools.session.ServerSession;

/**
 * {@link AppPasswordApplicationsConverter}
 * Convert an Application List Result to JSON for output
 *
 * @author <a href="mailto:greg.hill@open-xchange.com">Greg Hill</a>
 * @since v7.10.4
 */
public class AppPasswordApplicationsConverter implements ResultConverter {

    public static final String INPUT_FORMAT = "AppList";
    private static final String OUTPUT_FORMAT = "json";

    /**
     * Initializes a new {@link AppPasswordApplicationsConverter}.
     */
    public AppPasswordApplicationsConverter() {
        super();
    }

    @Override
    public String getInputFormat() {
        return INPUT_FORMAT;
    }

    @Override
    public String getOutputFormat() {
        return OUTPUT_FORMAT;
    }

    @Override
    public Quality getQuality() {
        return Quality.GOOD;
    }

    @SuppressWarnings("unused")
    @Override
    public void convert(AJAXRequestData requestData, AJAXRequestResult result, ServerSession session, Converter converter) throws OXException {
        Object resultObject = result.getResultObject();
        if (!(resultObject instanceof List)) {
            throw new UnsupportedOperationException();
        }
        @SuppressWarnings("unchecked") List<Object> apps = (List<Object>) resultObject;
        try {
            JSONArray json = new JSONArray(apps.size());
            for (Object app : apps) {
                json.put(toJSON((AppPasswordApplication) app));
            }
            result.setResultObject(json);
        } catch (JSONException e) {
            throw OXJSONExceptionCodes.JSON_WRITE_ERROR.create(e);
        }
    }

    /**
     * Simple mapping of applications
     *
     * @param app The AppPasswordApplication
     * @return JSON representation
     * @throws JSONException if a JSON error is occurred
     */
    private static JSONObject toJSON(AppPasswordApplication app) throws JSONException {
        JSONObject result = new JSONObject();
        result.put("name", app.getType());
        result.put("displayName", app.getDisplayName());
        result.putSafe("sort", app.getSortOrder());
        return result;
    }
}
